/* -*- C++ -*-
 * src/lemon/array_map.h - Part of LEMON, a generic C++ optimization library
 *
 * Copyright (C) 2004 Egervary Jeno Kombinatorikus Optimalizalasi Kutatocsoport
 * (Egervary Combinatorial Optimization Research Group, EGRES).
 *
 * Permission to use, modify and distribute this software is granted
 * provided that this copyright notice appears in all copies. For
 * precise terms see the accompanying LICENSE file.
 *
 * This software is provided "AS IS" with no warranty of any kind,
 * express or implied, and with no claim as to its suitability for any
 * purpose.
 *
 */

#ifndef LEMON_ARRAY_MAP_H
#define LEMON_ARRAY_MAP_H

#include <memory>

#include <lemon/map_iterator.h>

///\ingroup graphmaps
///\file
///\brief Graph maps that construates and destruates
///their elements dynamically.

namespace lemon {


  /// \addtogroup graphmaps
  /// @{
	
  /** The ArrayMap template class is graph map structure what
   *  automatically updates the map when a key is added to or erased from
   *  the map. This map factory uses the allocators to implement 
   *  the container functionality.
   *
   *  The template parameter is the MapRegistry that the maps
   *  will belong to and the ValueType.
   */

  template <typename _Graph, 
	    typename _Item,
	    typename _ItemIt,
	    typename _IdMap,
	    typename _Value>
  class ArrayMap : public AlterationObserverRegistry<_Item>::ObserverBase {

  public:
		
    /// The graph type of the maps. 
    typedef _Graph Graph;
    /// The key type of the maps.
    typedef _Item KeyType;

    typedef AlterationObserverRegistry<_Item> Registry;

  private:
    /// The iterator to iterate on the keys.
    typedef _ItemIt KeyIt;

    typedef _IdMap IdMap;

    typedef _Value Value;

    /// The MapBase of the Map which imlements the core regisitry function.
    typedef typename Registry::ObserverBase Parent;
		
    
  public:

    /// The value type of the map.
    typedef Value ValueType;
    /// The reference type of the map;
    typedef Value& ReferenceType;
    /// The pointer type of the map;
    typedef Value* PointerType;

    /// The const value type of the map.
    typedef const Value ConstValueType;
    /// The const reference type of the map;
    typedef const Value& ConstReferenceType;
    /// The pointer type of the map;
    typedef const Value* ConstPointerType;


  private:
    typedef std::allocator<Value> Allocator;


  public:

    /** Graph and Registry initialized map constructor.
     */
    ArrayMap(const Graph& _g, Registry& _r) : graph(&_g) {
      attach(_r);
      allocate_memory();
      for (KeyIt it(*graph); it != INVALID; ++it) {
	int id = IdMap(*graph)[it];
	allocator.construct(&(values[id]), Value());
      }								
    }

    /// Constructor to use default value to initialize the map. 

    /// It constrates a map and initialize all of the the map. 

    ArrayMap(const Graph& _g, Registry& _r, const Value& _v) : graph(&_g) {
      attach(_r);
      allocate_memory();
      for (KeyIt it(*graph); it != INVALID; ++it) {
	int id = IdMap(*graph)[it];
	allocator.construct(&(values[id]), _v);
      }								
    }

    /** Constructor to copy a map of the same map type.
     */
    ArrayMap(const ArrayMap& copy) {
      if (copy.attached()) {
	attach(*copy.getRegistry());
      }
      capacity = copy.capacity;
      if (capacity == 0) return;
      values = allocator.allocate(capacity);
      for (KeyIt it(*graph); it != INVALID; ++it) {
	int id = IdMap(*graph)[it];
	allocator.construct(&(values[id]), copy.values[id]);
      }
    }

    /** Assign operator to copy a map of the same map type.
     */
    ArrayMap& operator=(const ArrayMap& copy) {
      if (&copy == this) return *this;
      
      if (graph != copy.graph) {
	if (attached()) {
	  clear();
	  detach();
	}
	if (copy.attached()) {
	  attach(*copy.getRegistry());
	}
	capacity = copy.capacity;
	if (capacity == 0) return *this;
	values = allocator.allocate(capacity);      
      }

      for (KeyIt it(*graph); it != INVALID; ++it) {
	int id = IdMap(*graph)[it];
	allocator.construct(&(values[id]), copy.values[id]);
      }

      return *this;
    }

    /** The destructor of the map.
     */
    virtual ~ArrayMap() {      
      if (attached()) {
	clear();
	detach();
      }
    }
	
	
    /**
     * The subscript operator. The map can be subscripted by the
     * actual keys of the graph. 
     */
    ReferenceType operator[](const KeyType& key) {
      int id = IdMap(*graph)[key];
      return values[id];
    } 
		
    /**
     * The const subscript operator. The map can be subscripted by the
     * actual keys of the graph. 
     */
    ConstReferenceType operator[](const KeyType& key) const {
      int id = IdMap(*graph)[key];
      return values[id];
    }
	
    /** Setter function of the map. Equivalent with map[key] = val.
     *  This is a compatibility feature with the not dereferable maps.
     */
    void set(const KeyType& key, const ValueType& val) {
      (*this)[key] = val;
    }
		
    /** Add a new key to the map. It called by the map registry.
     */
    void add(const KeyType& key) {
      int id = IdMap(*graph)[key];
      if (id >= capacity) {
	int new_capacity = (capacity == 0 ? 1 : capacity);
	while (new_capacity <= id) {
	  new_capacity <<= 1;
	}
	Value* new_values = allocator.allocate(new_capacity);
	for (KeyIt it(*graph); it != INVALID; ++it) {
	  int jd = IdMap(*graph)[it];
	  if (id != jd) {
	    allocator.construct(&(new_values[jd]), values[jd]);
	    allocator.destroy(&(values[jd]));
	  }
	}
	if (capacity != 0) allocator.deallocate(values, capacity);
	values = new_values;
	capacity = new_capacity;
      }
      allocator.construct(&(values[id]), Value());
    }
		
    /** Erase a key from the map. It called by the map registry.
     */
    void erase(const KeyType& key) {
      int id = IdMap(*graph)[key];
      allocator.destroy(&(values[id]));
    }

    void build() {
      allocate_memory();
      for (KeyIt it(*graph); it != INVALID; ++it) {
	int id = IdMap(*graph)[it];
	allocator.construct(&(values[id]), Value());
      }								
    }

    void clear() {	
      if (capacity != 0) {
	for (KeyIt it(*graph); it != INVALID; ++it) {
	  int id = IdMap(*graph)[it];
	  allocator.destroy(&(values[id]));
	}								
	allocator.deallocate(values, capacity);
	capacity = 0;
      }
    }

//     /// The stl compatible pair iterator of the map.
//     typedef MapIterator<ArrayMap> Iterator;
//     /// The stl compatible const pair iterator of the map.
//     typedef MapConstIterator<ArrayMap> ConstIterator;

//     /** Returns the begin iterator of the map.
//      */
//     Iterator begin() {
//       return Iterator(*this, KeyIt(*MapBase::getGraph()));
//     }

//     /** Returns the end iterator of the map.
//      */
//     Iterator end() {
//       return Iterator(*this, INVALID);
//     }

//     /** Returns the begin ConstIterator of the map.
//      */
//     ConstIterator begin() const {
//       return ConstIterator(*this, KeyIt(*MapBase::getGraph()));
//     }

//     /** Returns the end const_iterator of the map.
//      */
//     ConstIterator end() const {
//       return ConstIterator(*this, INVALID);
//     }

//     /// The KeySet of the Map.
//     typedef MapConstKeySet<ArrayMap> ConstKeySet;

//     /// KeySet getter function.
//     ConstKeySet keySet() const {
//       return ConstKeySet(*this); 
//     }

//     /// The ConstValueSet of the Map.
//     typedef MapConstValueSet<ArrayMap> ConstValueSet;

//     /// ConstValueSet getter function.
//     ConstValueSet valueSet() const {
//       return ConstValueSet(*this);
//     }

//     /// The ValueSet of the Map.
//     typedef MapValueSet<ArrayMap> ValueSet;

//     /// ValueSet getter function.
//     ValueSet valueSet() {
//       return ValueSet(*this);
//     }

  private:
      
    void allocate_memory() {
      int max_id = IdMap(*graph).maxId();
      if (max_id == -1) {
	capacity = 0;
	values = 0;
	return;
      }
      capacity = 1;
      while (capacity <= max_id) {
	capacity <<= 1;
      }
      values = allocator.allocate(capacity);	
    }      

    const Graph* graph;
    int capacity;
    Value* values;
    Allocator allocator;

  public:
//     // STL  compatibility typedefs.
//     typedef Iterator iterator;
//     typedef ConstIterator const_iterator;
//     typedef typename Iterator::PairValueType value_type;
//     typedef typename Iterator::KeyType key_type;
//     typedef typename Iterator::ValueType data_type;
//     typedef typename Iterator::PairReferenceType reference;
//     typedef typename Iterator::PairPointerType pointer;
//     typedef typename ConstIterator::PairReferenceType const_reference;
//     typedef typename ConstIterator::PairPointerType const_pointer;
//     typedef int difference_type;		
  };		

  template <typename _Base> 
  class ArrayMappableGraphExtender : public _Base {
  public:

    typedef ArrayMappableGraphExtender<_Base> Graph;
    typedef _Base Parent;

    typedef typename Parent::Node Node;
    typedef typename Parent::NodeIt NodeIt;
    typedef typename Parent::NodeIdMap NodeIdMap;
    typedef typename Parent::NodeObserverRegistry NodeObserverRegistry;

    typedef typename Parent::Edge Edge;
    typedef typename Parent::EdgeIt EdgeIt;
    typedef typename Parent::EdgeIdMap EdgeIdMap;
    typedef typename Parent::EdgeObserverRegistry EdgeObserverRegistry;

    

    template <typename _Value>
    class NodeMap : public ArrayMap<Graph, Node, NodeIt, NodeIdMap, _Value> {
    public:
      typedef ArrayMappableGraphExtender<_Base> Graph;

      typedef typename Graph::Node Node;
      typedef typename Graph::NodeIt NodeIt;
      typedef typename Graph::NodeIdMap NodeIdMap;

      typedef ArrayMap<Graph, Node, NodeIt, NodeIdMap, _Value> Parent;

      typedef typename Parent::Graph Graph;
      typedef typename Parent::Value Value;

      NodeMap(const Graph& g) 
	: Parent(g, g.getNodeObserverRegistry()) {}
      NodeMap(const Graph& g, const Value& v) 
	: Parent(g, g.getNodeObserverRegistry(), v) {}

    };

    template <typename _Value>
    class EdgeMap : public ArrayMap<Graph, Edge, EdgeIt, EdgeIdMap, _Value> {
    public:
      typedef ArrayMappableGraphExtender<_Base> Graph;

      typedef typename Graph::Edge Edge;
      typedef typename Graph::EdgeIt EdgeIt;
      typedef typename Graph::EdgeIdMap EdgeIdMap;

      typedef ArrayMap<Graph, Edge, EdgeIt, EdgeIdMap, _Value> Parent;

      typedef typename Parent::Graph Graph;
      typedef typename Parent::Value Value;

      EdgeMap(const Graph& g) 
	: Parent(g, g.getEdgeObserverRegistry()) {}
      EdgeMap(const Graph& g, const Value& v) 
	: Parent(g, g.getEdgeObserverRegistry(), v) {}

    };
    
  };

/// @}

}

#endif //LEMON_ARRAY_MAP_H
